@@ -1,11 +1,17 @@
 #!/usr/bin/env python3
 
-"""Provide information about CI artifacts and make them available locally"""
+"""Provide information about CI artifacts and make them available locally
+Copyright (C) 2024 Checkmk GmbH - License: GNU General Public License v2
+This file is part of Checkmk (https://checkmk.com). It is subject to the terms and
+conditions defined in the file COPYING, which is part of this source code package.
+"""
 
 # pylint: disable=too-many-branches
 # pylint: disable=too-many-arguments
 # pylint: disable=fixme
 
+import asyncio
+import json
 import logging
 import os
 import sys
@@ -18,12 +24,14 @@ from datetime import datetime
 from itertools import chain
 from pathlib import Path
 from subprocess import check_output
-from typing import Any, cast
+from typing import Any, Literal, cast
 
+import requests
+from jenkins import Jenkins
 from trickkiste.logging_helper import apply_common_logging_cli_args, setup_logging
 from trickkiste.misc import compact_dict, cwd, md5from, split_params
 
-from cmk_dev.jenkins_utils import (
+from .jenkins_utils import (
     AugmentedJenkinsClient,
     Build,
     BuildId,
@@ -36,14 +44,15 @@ from cmk_dev.jenkins_utils import (
     extract_credentials,
     params_from,
 )
-from cmk_dev.utils import Fatal
-from jenkins import Jenkins
+from .utils import Fatal
+from .version import __version__
 
 # Todo: warn about missing parameters
 # Todo: default to `$REPO/package_download` rather than `out`
 # Todo: decent error message when providing wrong path
 
 PathHashes = Mapping[str, str]
+shared_build_info: asyncio.Queue[str] = asyncio.Queue(maxsize=1)
 
 
 def parse_args() -> Args:
@@ -56,6 +65,8 @@ def parse_args() -> Args:
     parser.set_defaults(func=lambda *_: parser.print_usage())
     subparsers = parser.add_subparsers(help="available commands", metavar="CMD")
 
+    parser.add_argument("--version", action="version", version=__version__)
+
     parser_info = subparsers.add_parser("info")
     parser_info.set_defaults(func=_fn_info)
     parser_info.add_argument(
@@ -110,12 +121,6 @@ def parse_args() -> Args:
             action="store_true",
             help="Don't check for existing matching builds, instead start a new build immediately",
         )
-        subparser.add_argument(
-            "-n",
-            "--omit-new-build",
-            action="store_true",
-            help="Don't issue new builds, even if no matching build could be found",
-        )
 
     def apply_download_args(subparser: ArgumentParser) -> None:
         subparser.add_argument(
@@ -131,34 +136,67 @@ def parse_args() -> Args:
             help="If set, existing files not part of artifacts won't be deleted",
         )
 
-    parser_request = subparsers.add_parser("request", help="Request a build")
+    parser_request = subparsers.add_parser(
+        "request", help="Request a build or identify an existing one."
+    )
     parser_request.set_defaults(func=_fn_request_build)
     apply_common_args(parser_request)
     apply_request_args(parser_request)
+    parser_request.add_argument(
+        "--passive",
+        action="store_true",
+        help=(
+            "Instead of actually triggering new builds, write out the information needed to"
+            " trigger a build instead. To be used inside pipline scripts in order to keep track"
+            " of issuer."
+        ),
+    )
 
-    parser_await_result = subparsers.add_parser("await-result")
+    parser_await_result = subparsers.add_parser(
+        "await-result", help="Wait for existing build to finish, don't download."
+    )
     parser_await_result.set_defaults(func=_fn_await_and_handle_build, download=False)
     apply_common_args(parser_await_result)
     parser_await_result.add_argument("build_number", type=int, nargs="?")
 
-    parser_download = subparsers.add_parser("download")
+    parser_download = subparsers.add_parser(
+        "download",
+        help="Download artifacts of existing build. Wait for the build to finish if necessary.",
+    )
     parser_download.set_defaults(func=_fn_await_and_handle_build, download=True)
     apply_common_args(parser_download)
     apply_download_args(parser_download)
     parser_download.add_argument("build_number", type=int, nargs="?")
 
-    parser_fetch = subparsers.add_parser("fetch")
-    parser_fetch.set_defaults(func=_fn_fetch)
+    parser_fetch = subparsers.add_parser(
+        "fetch",
+        help="Trigger or identify matching build, wait for it to finish, download artifacts.",
+    )
+    parser_fetch.set_defaults(func=_fn_fetch, download=True)
     apply_common_args(parser_fetch)
     apply_request_args(parser_fetch)
     apply_download_args(parser_fetch)
+    parser_fetch.add_argument(
+        "-n",
+        "--omit-new-build",
+        action="store_true",
+        help="Don't issue new builds, even if no matching build could be found",
+    )
+
+    parser_validate = subparsers.add_parser(
+        "validate", help="Request, identify matching or trigger build and wait for it to finish."
+    )
+    parser_validate.set_defaults(func=_fn_fetch, download=False, omit_new_build=False)
+    apply_common_args(parser_validate)
+    apply_request_args(parser_validate)
+    apply_download_args(parser_validate)
 
     return parser.parse_args()
 
 
 def log() -> logging.Logger:
     """Convenience function retrieves 'our' logger"""
-    return logging.getLogger("cmk-dev.cia")
+    return logging.getLogger("trickkiste.cmk-dev.cia")
 
 
 def flatten(params: None | Sequence[JobParams]) -> None | JobParams:
@@ -166,16 +204,17 @@ def flatten(params: None | Sequence[JobParams]) -> None | JobParams:
     return {key: value for param in params for key, value in param.items()} if params else None
 
 
-def _fn_info(args: Args) -> None:
+async def _fn_info(args: Args) -> None:
     """Entry point for information about job artifacts"""
-    with AugmentedJenkinsClient(
+    async with AugmentedJenkinsClient(
         **extract_credentials(args.credentials), timeout=args.timeout
     ) as jenkins:
-        class_name = (job_info := jenkins.raw_job_info(args.job))["_class"]
+        job_info = await jenkins.raw_job_info(args.job)
+        class_name = job_info["_class"]
         if class_name == "com.cloudbees.hudson.plugins.folder.Folder":
             print(f"Folder({job_info['name']}, jobs: {len(cast(list[Any], job_info['jobs']))})")
         elif class_name == "org.jenkinsci.plugins.workflow.job.WorkflowJob":
-            job = Job.model_validate(job_info).expand(jenkins)
+            job = await Job.model_validate(job_info).expand(jenkins)
             print(job)
             for build_nr, build in job.build_infos.items():
                 print(f"  - {build_nr}: {build}")
@@ -276,14 +315,28 @@ def download_artifacts(
                 fp_hash,
             )
 
-        with client._session.get(f"{build.url}artifact/{artifact}", stream=True) as reply:
-            log().debug("download: %s", artifact)
-            reply.raise_for_status()
-            artifact_filename.parent.mkdir(parents=True, exist_ok=True)
-            with open(artifact_filename, "wb") as out_file:
-                for chunk in reply.iter_content(chunk_size=1 << 16):
-                    out_file.write(chunk)
-            downloaded_artifacts.append(artifact)
+        MAX_RETRIES = 3
+        for attempt in range(MAX_RETRIES):
+            try:
+                with client._session.get(f"{build.url}artifact/{artifact}", stream=True) as reply:
+                    log().debug("download: %s", artifact)
+                    reply.raise_for_status()
+                    artifact_filename.parent.mkdir(parents=True, exist_ok=True)
+                    with open(artifact_filename, "wb") as out_file:
+                        for chunk in reply.iter_content(chunk_size=8192):
+                            if chunk:  # Filter out keep-alive chunks
+                                out_file.write(chunk)
+                    downloaded_artifacts.append(artifact)
+                break
+            except requests.exceptions.ChunkedEncodingError as e:
+                log().info("Retrying due to chunked encoding error: %s (attempt %d/%d)", e, attempt + 1, MAX_RETRIES)
+                if attempt == MAX_RETRIES - 1:
+                    raise
+            except requests.exceptions.ConnectionError as e:
+                # like "Remote end closed connection without response"
+                log().info("Retrying due to connection error error: %s (attempt %d/%d)", e, attempt + 1, MAX_RETRIES)
+                if attempt == MAX_RETRIES - 1:
+                    raise
 
     if not no_remove_others:
         for path in existing_files - set(downloaded_artifacts) - set(skipped_artifacts):
@@ -330,11 +383,25 @@ def find_mismatching_parameters(
 ) -> Sequence[tuple[str, JobParamValue, JobParamValue]]:
     """Returns list of key and mismatching values in mapping @first which also occur in @second"""
     # TODO: find solution for unprovided parameters and default/empty values
-    return [
-        (key, cast(JobParamValue, first.get(key, "")), cast(JobParamValue, second.get(key, "")))
-        for key in set(first.keys() | second.keys()) - {"DISABLE_CACHE"}
-        if first.get(key) and first.get(key, "") != second.get(key, "")
-    ]
+    bool_map = {"true": True, "false": False}
+    mismatching_parameters = []
+    for key in set(first.keys() | second.keys()) - {"DISABLE_CACHE"}:
+        if first_val := first.get(key, ""):
+            second_val = second.get(key, "")
+
+            assert first_val not in ["TRUE", "FALSE", "True", "False"], "Only groovy lower case bool 'false' or 'true' allowed"
+
+            # apply bool mapping in most stupid way
+            if isinstance(first_val, str):
+                first_val = bool_map.get(first_val, first_val)
+            if isinstance(second_val, str):
+                second_val = bool_map.get(second_val, second_val)
+
+            if first_val != second_val:
+                mismatching_parameters.append(
+                    (key, cast(JobParamValue, first_val), cast(JobParamValue, second_val))
+                )
+    return mismatching_parameters
 
 
 def meets_constraints(
@@ -395,15 +462,14 @@ def meets_constraints(
                 path_hashes,
             )
             result = False
-    else:
-        if not path_hashes_match(expected_path_hashes, path_hashes):
-            log().debug(
-                "build #%s has been started with mismatching expected path hashes: %s != %s",
-                build.number,
-                expected_path_hashes,
-                path_hashes,
-            )
-            result = False
+    elif not path_hashes_match(expected_path_hashes, path_hashes):
+        log().debug(
+            "build #%s has been started with mismatching expected path hashes: %s != %s",
+            build.number,
+            expected_path_hashes,
+            path_hashes,
+        )
+        result = False
 
     if time_constraints is None:
         pass
@@ -447,14 +513,14 @@ def build_id_from_queue_item(client: Jenkins, queue_id: QueueId) -> BuildId:
         time.sleep(1)
 
 
-def find_matching_queue_item(
+async def find_matching_queue_item(
     jenkins_client: AugmentedJenkinsClient,
     job: Job,
     params: None | JobParams,
     path_hashes: PathHashes,
 ) -> None | BuildId:
     """Looks for a queued build matching job and parameters and returns the QueueId"""
-    for queue_item in jenkins_client.client.get_queue_info():
+    for queue_item in await jenkins_client.queue_info():
         if not cast(str, queue_item.get("_class", "")).startswith("hudson.model.Queue"):
             continue
         if cast(str, cast(GenMap, queue_item.get("task", {})).get("url", "")) != job.url:
@@ -513,28 +579,80 @@ def compose_out_dir(base_dir: Path, out_dir: Path) -> Path:
     return out_dir
 
 
-def _fn_request_build(args: Args) -> None:
-    """Entry point for a build request"""
-    with AugmentedJenkinsClient(
+async def _fn_request_build(args: Args) -> None:
+    """Entry point for a build request
+    If none of the existing builds match the conditions a new build will be
+    issued.
+    This can get complicated since we don't know the outcome of unfinished or
+    queued elements yet (result and dependency path hashes).
+    """
+    async with AugmentedJenkinsClient(
         **extract_credentials(args.credentials), timeout=args.timeout
     ) as jenkins_client:
-        if not (job := jenkins_client.job_info(args.job)).type == "WorkflowJob":
+        if not (job := await jenkins_client.job_info(args.job)).type == "WorkflowJob":
             raise Fatal(f"{args.job} is not a WorkflowJob")
+        # In case we force a new build anyway we don't have to look for an existing one
+        if matching_build := (
+            None
+            if args.force_new_build
+            else await identify_matching_build(
+                job,
+                jenkins_client=jenkins_client,
+                params=flatten(args.params),
+                path_hashes=compose_path_hashes(args.base_dir, args.dependency_paths),
+                time_constraints=args.time_constraints,
+            )
+        ):
+            print(
+                json.dumps(
+                    {
+                        "existing": {
+                            "path": job.path,
+                            "number": matching_build.number,
+                            "url": matching_build.url,
+                            "result": matching_build.result,
+                            "new_build": False,
+                            # "parameters": matching_build.parameters,
+                            # "path_hashes": matching_build.path_hashes,
+                        }
+                    }
+                )
+            )
+            return
 
-        build_candidate = request_matching_build(
-            job,
-            jenkins_client=jenkins_client,
+        new_build_params = compose_build_params(
             params=flatten(args.params),
             params_no_check=flatten(args.params_no_check),
             path_hashes=compose_path_hashes(args.base_dir, args.dependency_paths),
-            time_constraints=args.time_constraints,
-            omit_new_build=args.omit_new_build,
-            force_new_build=args.force_new_build,
         )
-        print(f"{job.path}:{build_candidate.number}:{build_candidate.url}")
+        if args.passive:
+            print(
+                json.dumps(
+                    {
+                        "new_build": {
+                            "path": job.path,
+                            "params": new_build_params,
+                        }
+                    }
+                )
+            )
+        else:
+            new_build = await trigger_build(jenkins_client, job, new_build_params)
+            print(
+                json.dumps(
+                    {
+                        "triggered_build": {
+                            "path": job.path,
+                            "number": new_build.number,
+                            "url": new_build.url,
+                            "params": new_build_params,
+                        }
+                    }
+                )
+            )
 
 
-def _fn_await_and_handle_build(args: Args) -> None:
+async def _fn_await_and_handle_build(args: Args) -> None:
     """Entry point for artifacts download only"""
     out_dir = args.base_dir / (getattr(args, "out_dir", "") or "")
     if out_dir.exists() and not out_dir.is_dir():
@@ -548,122 +666,167 @@ def _fn_await_and_handle_build(args: Args) -> None:
         raise Fatal("Provide only one of separate build number or composite build name")
 
     if not job_number:
-        raise Fatal("No build number provided. Use either --build-number or `<job-name>:<number>`.")
+        raise Fatal("No build number provided. Use `<job-name>:<number>`.")
 
-    with AugmentedJenkinsClient(
+    async with AugmentedJenkinsClient(
         **extract_credentials(args.credentials), timeout=args.timeout
     ) as jenkins_client:
-        completed_build = await_build(
+        completed_build = await await_build(
             job_name,
             job_number,
             jenkins_client=jenkins_client,
             check_result=True,
             path_hashes=None,
+            allow_to_cancel=False,
+        )
+        print(
+            json.dumps(
+                {
+                    "result": completed_build.result,
+                    "artifacts": (
+                        list(
+                            chain(
+                                *download_artifacts(
+                                    jenkins_client.client,
+                                    completed_build,
+                                    out_dir,
+                                    args.no_remove_others,
+                                )
+                            )
+                        )
+                        if args.download
+                        else None
+                    ),
+                }
+            )
         )
-        if args.download:
-            for artifact in chain(
-                *download_artifacts(
-                    jenkins_client.client,
-                    completed_build,
-                    out_dir,
-                    args.no_remove_others,
-                )
-            ):
-                print(artifact)
-        else:
-            print(f"Build was {completed_build.result}")
 
 
-def _fn_fetch(args: Args) -> None:
+async def _fn_fetch(args: Args) -> None:
     """Entry point for fetching (request and download combined) artifacts"""
     out_dir = compose_out_dir(args.base_dir, args.out_dir)
     path_hashes = compose_path_hashes(args.base_dir, args.dependency_paths)
-    with AugmentedJenkinsClient(
+    async with AugmentedJenkinsClient(
         **extract_credentials(args.credentials), timeout=args.timeout
     ) as jenkins_client:
-        if not (job := jenkins_client.job_info(args.job)).type == "WorkflowJob":
+        if not (job := await jenkins_client.job_info(args.job)).type == "WorkflowJob":
             raise Fatal(f"{args.job} is not a WorkflowJob")
+        # In case we force a new build anyway we don't have to look for an existing one
+        matching_build = (
+            None
+            if args.force_new_build
+            else await identify_matching_build(
+                job,
+                jenkins_client=jenkins_client,
+                params=flatten(args.params),
+                path_hashes=compose_path_hashes(args.base_dir, args.dependency_paths),
+                time_constraints=args.time_constraints,
+            )
+        )
+        if args.omit_new_build and not matching_build:
+            raise Fatal(f"No matching build found for job '{job.name}' but new builds are omitted.")
 
-        build_candidate = request_matching_build(
+        build_candidate = matching_build or await trigger_build(
+            jenkins_client,
             job,
-            jenkins_client=jenkins_client,
-            params=flatten(args.params),
-            params_no_check=flatten(args.params_no_check),
-            path_hashes=path_hashes,
-            time_constraints=args.time_constraints,
-            omit_new_build=args.omit_new_build,
-            force_new_build=args.force_new_build,
+            compose_build_params(
+                params=flatten(args.params),
+                params_no_check=flatten(args.params_no_check),
+                path_hashes=compose_path_hashes(args.base_dir, args.dependency_paths),
+            ),
         )
 
         for key, value in build_candidate.__dict__.items():
             log().debug("  %s: %s", key, value)
 
-        completed_build = await_build(
+        completed_build = await await_build(
             job.path,
             build_candidate.number,
             jenkins_client=jenkins_client,
             check_result=True,
             path_hashes=path_hashes,
         )
-
-        for artifact in chain(
-            *download_artifacts(
-                jenkins_client.client,
-                completed_build,
-                out_dir,
-                args.no_remove_others,
+        downloaded_artifacts = (
+            list(
+                chain(
+                    *download_artifacts(
+                        jenkins_client.client,
+                        completed_build,
+                        out_dir,
+                        args.no_remove_others,
+                    )
+                )
             )
-        ):
-            print(artifact)
+            if args.download
+            else []
+        )
+
+        print(
+            json.dumps(
+                {
+                    "result": completed_build.result,
+                    "artifacts": downloaded_artifacts,
+                }
+            )
+        )
 
 
-def request_matching_build(
+async def identify_matching_build(
     job: Job,
     *,
     jenkins_client: AugmentedJenkinsClient,
     params: None | JobParams,
-    params_no_check: None | JobParams,
     path_hashes: PathHashes,
     time_constraints: None | str,
-    omit_new_build: bool,
-    force_new_build: bool,
-) -> Build:
+) -> None | Build:
     """Find an existing build (finished, still running or queued) which matches our
     requirements specified by @job_full_path matching @params and
     @time_constraints.
-    If none of the existing builds match the conditions a new build will be
-    issued.
-    This can get complicated since we don't know the outcome of unfinished or
-    queued elements yet (result and dependency path hashes).
     """
     # pylint: disable=too-many-locals
 
-    # In case we force a new build anyway we don't have to look for an existing one
-    if not force_new_build:
-        # fetch a job's build history first
-        job.expand(jenkins_client)
+    # fetch a job's build history first
+    await job.expand(jenkins_client)
 
-        # Look for finished builds
-        for build in filter(lambda b: b.completed, job.build_infos.values()):
-            if meets_constraints(build, params, time_constraints, path_hashes):
-                log().info("found matching finished build: %s (%s)", build.number, build.url)
-                return build
+    # Look for finished builds
+    for build in filter(lambda b: b.completed, job.build_infos.values()):
+        if meets_constraints(build, params, time_constraints, path_hashes):
+            log().info("found matching finished build: %s (%s)", build.number, build.url)
+            return build
 
-        # Look for still unfinished builds
-        for build in filter(lambda b: not b.completed, job.build_infos.values()):
-            if meets_constraints(build, params, time_constraints, path_hashes):
-                log().info("found matching unfinished build: %s (%s)", build.number, build.url)
-                return build
+    # Look for still unfinished builds
+    for build in filter(lambda b: not b.completed, job.build_infos.values()):
+        if meets_constraints(build, params, time_constraints, path_hashes):
+            log().info("found matching unfinished build: %s (%s)", build.number, build.url)
+            return build
 
-        if matching_item := find_matching_queue_item(jenkins_client, job, params, path_hashes):
-            return jenkins_client.build_info(job.path, matching_item)
+    if matching_item := await find_matching_queue_item(jenkins_client, job, params, path_hashes):
+        return await jenkins_client.build_info(job.path, matching_item)
 
-    if omit_new_build:
-        raise Fatal(f"No matching build found for job '{job.name}' but new builds are omitted.")
+    return None
 
-    parameters = {
-        **(params or {}),
-        **(params_no_check or {}),
+def convert_params(params: JobParams) -> JobParams:
+    """convert params to use real boolean True/False instead of 'true'/'false'"""
+    bool_map = {"true": True, "false": False}
+    converted_params: JobParams = {}
+    for key, val in params.items():
+        # apply bool mapping in most stupid way
+        if isinstance(val, str):
+            converted_params[key] = bool_map.get(val, val)
+        else:
+            converted_params[key] = val
+
+    return converted_params
+
+def compose_build_params(
+    params: None | JobParams,
+    params_no_check: None | JobParams,
+    path_hashes: PathHashes,
+) -> JobParams:
+    """Convenience function combining job parameters"""
+    return {
+        **(convert_params(params=params or {})),
+        **(convert_params(params=params_no_check or {})),
         **(
             {
                 "DEPENDENCY_PATH_HASHES": ",".join(
@@ -675,37 +838,47 @@ def request_matching_build(
         ),
     }
 
-    log().info("start new build for %s", job.path)
-    log().info("  params=%s", compact_dict(parameters))
 
-    return jenkins_client.build_info(
+async def trigger_build(
+    jenkins_client: AugmentedJenkinsClient,
+    job: Job,
+    params: JobParams,
+) -> Build:
+    """Convenience function triggering a build with given @params and waiting for build number"""
+    log().info("start new build for %s", job.path)
+    log().info("  params=%s", compact_dict(params))
+
+    return await jenkins_client.build_info(
         job.path,
         build_id_from_queue_item(
             jenkins_client.client,
-            jenkins_client.client.build_job(job.path, parameters=parameters),
+            jenkins_client.client.build_job(job.path, parameters=params),
         ),
     )
 
 
-def await_build(
+async def await_build(
     job_full_path: str,
     build_number: int,
     *,
     jenkins_client: AugmentedJenkinsClient,
     check_result: bool,
     path_hashes: None | PathHashes,
+    allow_to_cancel: bool = True,
 ) -> Build:
     """Awaits a Jenkins job build specified by @job_full_path and @build_number and returns the
     awaited Build object. Unexpected build failures or non-matching path hashes will be raised on.
     """
-    current_build_info = jenkins_client.build_info(job_full_path, build_number)
+    current_build_info = await jenkins_client.build_info(job_full_path, build_number)
     if not current_build_info.completed:
         log().info("build #%s still in progress (%s)", build_number, current_build_info.url)
+        if allow_to_cancel:
+            await shared_build_info.put(json.dumps({"path": job_full_path, "number": build_number}))
         while True:
             if not current_build_info.completed:
                 log().debug("build %s in progress", build_number)
                 time.sleep(10)
-                current_build_info = jenkins_client.build_info(job_full_path, build_number)
+                current_build_info = await jenkins_client.build_info(job_full_path, build_number)
                 continue
             break
 
@@ -726,6 +899,45 @@ def await_build(
     return current_build_info
 
 
+async def stop_build(args: Args, ongoing_build_info: Mapping[str, str | int]) -> None:
+    """Convenience function stopping a running or queued build"""
+    job_name = ongoing_build_info["path"]
+    job_number = ongoing_build_info["number"]
+    assert isinstance(job_name, str)
+    assert isinstance(job_number, int)
+
+    log().info("stop build %d of %s", job_number, job_name)
+
+    async with AugmentedJenkinsClient(
+        **extract_credentials(args.credentials), timeout=args.timeout
+    ) as jenkins_client:
+        jenkins_client.client.stop_build(name=job_name, number=job_number)
+
+
+def query_yes_no(
+    question: str = "Continue?", default: None | Literal["yes", "y", "ye", "no", "n"] = "no"
+) -> bool:
+    """Ask a yes/no question via raw_input() and return their answer."""
+    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
+    if default is None:
+        prompt = " [y/n] "
+    elif default == "yes":
+        prompt = " [Y/n] "
+    elif default == "no":
+        prompt = " [y/N] "
+    else:
+        raise ValueError(f"invalid default answer: '{default}'")
+
+    while True:
+        sys.stdout.write(question + prompt)
+        choice = input().lower()
+        if default is not None and choice == "":
+            return valid[default]
+        if choice in valid:
+            return valid[choice]
+        sys.stdout.write("Please respond with 'yes' or 'no' (or 'y' or 'n').\n")
+
+
 def main() -> None:
     """Entry point for everything else"""
     try:
@@ -737,10 +949,22 @@ def main() -> None:
             os.environ["FORCE_COLOR"] = "true"
             os.environ["COLUMNS"] = "200"
 
-        setup_logging(log(), args.log_level)
+        setup_logging(
+            logger=log(),
+            level=args.log_level,
+            show_time=False,
+            show_name=False,
+            show_funcname=False,
+        )
 
         log().debug("Parsed args: %s", args)
-        args.func(args)
+        asyncio.run(args.func(args))
+    except KeyboardInterrupt:
+        if not shared_build_info.empty():
+            if query_yes_no(question="Cancel ongoing build?"):
+                ongoing_build_info = json.loads(shared_build_info.get_nowait())
+                asyncio.run(stop_build(args=args, ongoing_build_info=ongoing_build_info))
+                log().debug("Stopping ongoing job")
     except Fatal as exc:
         print(exc, file=sys.stderr)
         raise SystemExit(-1) from exc
