@@ -1,6 +1,11 @@
 #!/usr/bin/env python3
 
-"""Prints ps output optionally filtered for CPU"""
+"""Prints ps output optionally filtered for CPU
+
+Copyright (C) 2024 Checkmk GmbH - License: GNU General Public License v2
+This file is part of Checkmk (https://checkmk.com). It is subject to the terms and
+conditions defined in the file COPYING, which is part of this source code package.
+"""
 
 import sys
 
